#include <algorithm>
#include "Level.hpp"
#include "player.hpp"
#include "Enemyprojectile.hpp"
#include "playerprojectile.hpp"
#include "PixelShaders.hpp"

// <kruegerzo72182 ----------------------------------------------------------------------
// Wall stuff --------------------------------------------------------------------------

DynamicPixelBuffer Wall::constructWallBuffer(Orientation orientation, int length, int thickness)
{
    if(orientation == Orientation::Horizontal) {
        return { length, thickness };
    }
    return { thickness, length };
}

Wall::Wall(int length, int thickness, Position const &pos, Orientation orientation) :
    Entity(pos),
    wallOrientation(orientation),
    wallBuffer(constructWallBuffer(orientation, length, thickness))
{
    this->wallBuffer.clear(_1);
}

PixelBuffer const& Wall::getSprite()
{
    return this->wallBuffer;
}

Orientation Wall::getOrientation() const
{
    return this->wallOrientation;
}

// Scenery stuff --------------------------------------------------------------------------

Scenery::Scenery(Position const &pos, SceneryType type)
    : Entity(pos), sType(type)
{
}

SceneryType Scenery::getSceneryType()
{
    return this->sType;
}

PixelBuffer const &Scenery::getSprite()
{
    // switch case returns depending on SceneryType here
    switch(sType) {
    case SceneryType::Dead_Tree: return this->DEAD_TREE; break;
    case SceneryType::Rock: return this->ROCK; break;
    case SceneryType::Rubble: return this->RUBBLE; break;
    case SceneryType::Bush: return this->BUSH; break;
    case SceneryType::Fail: return this->FAIL; break;
    case SceneryType::Win: return this->WIN; break;
    default:
        return this->FAIL;;
    }
}

// Level stuff --------------------------------------------------------------------------

Level::Level(int width, int heigth, Player &player) :
    levelWidth(width),
    levelHeigth(heigth),
    levelBuffer(width, heigth),
    player(player)
{
    int thickness = 10;
    borders.push_back(Wall(width, thickness, { 0.0f, 0.0f }, Orientation::Horizontal));
    borders.push_back(Wall(width, thickness, { 0.0f, static_cast<float>(heigth - thickness )}, Orientation::Horizontal));
    borders.push_back(Wall(heigth, thickness, { 0.0f, 0.0f }, Orientation::Vertical));
    borders.push_back(Wall(heigth, thickness, { static_cast<float>(width - thickness), 0.0f }, Orientation::Vertical));
}

void Level::addWall(int length, int thickness, Position const &topleft, Orientation orientation)
{
     if(std::none_of(
        wallVector.begin(), wallVector.end(), [&](Wall const &i) {
            return (
                (i.getTopLeft() == topleft) &&
                (orientation == i.getOrientation())
            );
    })) {
        wallVector.emplace_back(Wall{ length, thickness, topleft, orientation });
    }
}

void Level::addScenery(Position const &topleft, SceneryType type)
{
    if(std::none_of(
        sceneryVector.begin(), sceneryVector.end(), [&](Scenery const &i) {
            return (i.getTopLeft() == topleft);
    })) {
        sceneryVector.emplace_back(Scenery{ topleft, type });
    }
}

void Level::addEnemy(std::unique_ptr<Enemy> enemy_)
{
    enemyVector.push_back(std::move(enemy_));
}

// primes the trooperSpawner
void Level::fillTrooperSpawner(float posX, float posY, unsigned int nmbrEnemies)
{
    direction dir = direction::LEFT;
    for(unsigned int i = 0; i <= nmbrEnemies; ++i){
        trooperSpawner.push_back(std::make_unique<Trooper>(Trooper(posX,posY,dir,0)));
        dir=direction(rand()%4);
    }
}
// kruegerzo72182> ----------------------------------------------------------------------

// <hauerch71498 ----------------------------------------------------------------------
/**Enemy Addprojectile**/
void Level::addEnemyProjectile(EnemyProjectile const& projectile)
{
    projectileVector.push_back(std::make_unique<EnemyProjectile>(projectile));
};
// hauerch71498> ----------------------------------------------------------------------

// <weberma73121 ----------------------------------------------------------------------
/**Player Addprojectile**/
void Level::addPlayerProjectile(PlayerProjectile const& projectile)
{
    projectileVector.push_back(std::make_unique<PlayerProjectile>(projectile));
};
// weberma73121> ----------------------------------------------------------------------

// <kruegerzo72182 ----------------------------------------------------------------------
void Level::addTrooperFromSpawner()
{
    if(trooperSpawner.size() > 0){
        // Where is the std::vector pop_back() method that returns an rvalue
        // reference, dear C++ committee?! Where is it?!? --hoffmanncl72341
        this->enemyVector_PushBack(std::move(trooperSpawner.back()));
        trooperSpawner.pop_back();
    }
}

unsigned int Level::getEnemySpawnCount()
{
    return static_cast<unsigned int>(trooperSpawner.size());
}

void Level::enemyVector_PushBack(std::unique_ptr<Enemy> enemy_)
{
    this->enemyVector.push_back(std::move(enemy_));
}

// do all the Steps for movement and animations
bool Level::doDoSteps(FloatSeconds const &frame_time)
{
    player.step(player.getTopLeft(), frame_time, *this);
    if(player.getHitPoints() == 0) {
        return false;
    }

    for(auto it = enemyVector.begin(); it != enemyVector.end();) {
        auto& i = *it;
        if(i != nullptr && !i->step(player.getTopLeft(), frame_time, *this)) {
            it = enemyVector.erase(it);
        } else {
            it++;
        }
    }
// <hauerch71498 ----------------------------------------------------------------------
    for(auto it = projectileVector.begin(); it != projectileVector.end();){
        auto& i = *it;
        if(i != nullptr &&!i->doStep(frame_time, *this))
        {
            it = projectileVector.erase(it);
        }
        else
        {
            it++;
        }
    }
// hauerch71498> ----------------------------------------------------------------------
     for(auto it = sceneryVector.begin(); it != sceneryVector.end();) {
        auto& i = *it;
        if(!i.step(player.getTopLeft(), frame_time, *this)) {
            // call the step method from the Entity class to see if the object was hit
            it = sceneryVector.erase(it);
        } else {
            it++;
        }
    }

    for(auto it = wallVector.begin(); it != wallVector.end();) {
        auto& i = *it;
        if(!i.step(player.getTopLeft(), frame_time, *this)) {
            // call the step method from the Entity class to see if the object was hit
            it = wallVector.erase(it);
        } else {
            it++;
        }
    }
    return true;
}

// checks if the two given buffers are overlaping
bool Level::collHelper(Position const& vecPos, PixelBuffer const& vecBuf, float buffPosX, float buffPosY, int buffWidth, int buffHeigth)
{
    for(int x = static_cast<int>(vecPos.X); x <= static_cast<int>(vecPos.X) + vecBuf.w(); ++x){
        for(int y = static_cast<int>(vecPos.Y); y <= static_cast<int>(vecPos.Y) + vecBuf.h(); ++y){
            if((x >= buffPosX && x <= buffPosX + buffWidth)&&(y >= buffPosY && y <= buffPosY + buffHeigth)){
                return true;
            }
        }
    }
    return false;
}

// iterates through the border, walls and the scenery and checks for collisions
bool Level::checkCollision(float posX, float posY, int bufferWidth, int bufferHeigth)
{
    for(auto i = this->wallVector.begin(); i != this->wallVector.end(); i++ ){
        if(collHelper(i->getTopLeft(), i->getSprite(), posX, posY, bufferWidth, bufferHeigth)) return true;
    }
    for(auto i = this->borders.begin(); i != this->borders.end(); i++ ){
        if(collHelper(i->getTopLeft(), i->getSprite(), posX, posY, bufferWidth, bufferHeigth)) return true;
    }
    for(auto i = this->sceneryVector.begin(); i != this->sceneryVector.end(); i++ ){
        if(collHelper(i->getTopLeft(), i->getSprite(), posX, posY, bufferWidth, bufferHeigth)) return true;
    }
    return false;
}
// kruegerzo72182> ----------------------------------------------------------------------

// <hauerch71498 ----------------------------------------------------------------------
/*Hitdetection*/
bool Level::HitEnemy(float posX, float posY, int bufferWidth, int bufferHeigth)
{
    if(collHelper(player.getTopLeft(),player.getSprite(),posX,posY,bufferWidth,bufferHeigth))
    {
        player.takeDamage(ENEMY_PROJECTILE_COLOR);
        return true;
    }
    return false;
}

int Level::getScore() const
{
    return score;
}
// hauerch71498> ----------------------------------------------------------------------

// <weberma73121 ----------------------------------------------------------------------
// Hitdetection
bool Level::HitPlayer(float posX, float posY, int bufferWidth, int bufferHeigth)
{
    for(auto &i : this->enemyVector){
        if(i != nullptr && collHelper(i->getTopLeft(), i->getSprite(), posX, posY, bufferWidth, bufferHeigth))
        {
            if(i->takeDamage(PLAYER_PROJECTILE_COLOR) == 0)
            {
                score += i->getScore();
            }
            return true;
        }
    }
    return false;
}
/*For projectiles, Gameover*/
// weberma73121> ----------------------------------------------------------------------

// <kruegerzo72182 ----------------------------------------------------------------------
// remove scenery and wall objects from their corresponding vectors if theay are hit
// vector.erase() happens in doDoSteps() do to animation by Claudio
bool Level::HitObstacle(float posX, float posY, Color hitColor, int bufferWidth, int bufferHeigth)
{
    for(auto &i : wallVector){
        if(collHelper(i.getTopLeft(), i.getSprite(), posX, posY, bufferWidth, bufferHeigth))
        {
            i.takeDamage(hitColor);
            return true;
        }
    }
    for(auto &i : sceneryVector) {
        if(collHelper(i.getTopLeft(), i.getSprite(), posX, posY, bufferWidth, bufferHeigth))
        {
            i.takeDamage(hitColor);
            return true;
        }
    }
    return false;
}

// blit all objects, enemies and the player stored in the level class to the levelBuffer
DynamicPixelBuffer const& Level::draw()
{
    this->levelBuffer.clear(_t);

    for(auto i : this->wallVector){
        i.blit(this->levelBuffer);
    }
    for(auto i : this->borders){
        i.blit(this->levelBuffer);
    }
    for(auto i : this->sceneryVector){
        i.blit(this->levelBuffer);
    }
    this->player.blit(this->levelBuffer);
    for(auto & i : this->enemyVector){
        if(i != nullptr) {
            i->blit(this->levelBuffer);
        }
    }
// kruegerzo72182> ----------------------------------------------------------------------

// <hauerch71498 ----------------------------------------------------------------------
    for(auto & i : this->projectileVector){
        if(i != nullptr){
        this->levelBuffer.blit_topleft({ static_cast<int>(i->getX()), static_cast<int>(i->getY()) }, i->getSprite());}
    }
// hauerch71498> ----------------------------------------------------------------------
    return this->levelBuffer;
}
