#ifndef LEVEL_HPP
#define LEVEL_HPP

#include <vector>

#include "PixelBuffer.hpp"
#include "Trooper.hpp"

class EnemyProjectile;
class Projectile;
class PlayerProjectile;

class Enemy;
class Trooper;
class Player;

enum class Orientation
{
    Horizontal, Vertical
};

enum class SceneryType
{
    Dead_Tree, Rock, Rubble, Bush, Fail, Win
};

class Wall : public Entity
{
public:
    Wall(int length, int thickness, Position const &pos, Orientation orientation);
    DynamicPixelBuffer constructWallBuffer(Orientation orientation, int length, int thickness);
    PixelBuffer const& getSprite() override;
    Orientation getOrientation() const;

private:
    Orientation wallOrientation;
    DynamicPixelBuffer wallBuffer;
};

class Scenery : public Entity
{
public:
    Scenery(Position const &pos, SceneryType type);
    PixelBuffer const& getSprite() override;
    SceneryType getSceneryType();

private:
    SceneryType sType;

    static StaticPixelBuffer<32, 16> constexpr DEAD_TREE = {{{
        {"                  2     6       "},
        {"              2    2  62     2  "},
        {"     2   2  62      622   662   "},
        {"     2    2622  2 62222 662  2  "},
        {"      2    62    62226622  22   "},
        {"           622  226622          "},
        {"         2  62226222 22    6    "},
        {"     2    226222222      66222  "},
        {"      22    622222   666622     "},
        {"        22 6222666666222        "},
        {"          6222622222    222     "},
        {"      666222222     2           "},
        {"  222222662222       2          "},
        {"     62222622                   "},
        {"    22 62 26                    "},
        {"  2    2  6                     "}
    }}};
    static StaticPixelBuffer<8, 8> constexpr ROCK = {{{
        {"  FFFE  "},
        {" FE111E "},
        {"FE11111E"},
        {"F111FE11"},
        {"F11FE1E "},
        {"F11111E "},
        {" E11111E"},
        {" E11E1E "}
    }}};
    static StaticPixelBuffer<8, 8> constexpr BUSH = {{{
        {"   5545 "},
        {" 5444445"},
        {"54445444"},
        {" 4424545"},
        {"54244245"},
        {"  52445 "},
        {"  4224  "},
        {"  22222 "}
    }}};
    static StaticPixelBuffer<16, 8> constexpr RUBBLE = {{{
        {"      FE1       "},
        {"     F111E      "},
        {" FEE  1111   FE "},
        {"F111E  EE   FE1E"},
        {"F11FE1 11EE1F111"},
        {"F11111E11111F11E"},
        {" E11111E1F1F111E"},
        {" E11E1E F11111  "},
    }}};
    static StaticPixelBuffer<8, 8> constexpr FAIL = {{{
        {"3      3"},
        {"23    32"},
        {" 23  32 "},
        {"  2332  "},
        {"   32   "},
        {"  3223  "},
        {" 32  23 "},
        {"32    23"},
    }}};
    static StaticPixelBuffer<8, 8> constexpr WIN = {{{
        {"      54"},
        {"      54"},
        {"     54 "},
        {"     54 "},
        {"44  54  "},
        {"554 54  "},
        {"  544   "},
        {"   54   "},
    }}};
};

class Level
{
public:
    Level(int width, int heigth, Player &player);
    void addEnemyProjectile(EnemyProjectile const& projectile);
    void addPlayerProjectile(PlayerProjectile const& projectile);
    void addWall(int length, int thickness, Position const &topleft, Orientation orientation);
    void addScenery(Position const &topleft, SceneryType type);
    void addEnemy(std::unique_ptr<Enemy> enemy_);
    void fillTrooperSpawner(float posX, float posY, unsigned int nmbrEnemies);
    // Adds a new trooper from the trooper spawner, if any.
    void addTrooperFromSpawner();
    unsigned int getEnemySpawnCount();
    bool checkCollision(float posX, float posY, int bufferWidth, int bufferHeigth);
    // Did an enemy entity hit the player?
    bool HitEnemy(float posX, float posY, int bufferWidth, int bufferHeigth);
    // Did a player-controlled entity hit an Enemy?
    bool HitPlayer(float posX, float posY, int bufferWidth, int bufferHeigth);
    bool HitObstacle(float posX, float posY, Color hitColor, int bufferWidth, int bufferHeigth);
    void enemyVector_PushBack(std::unique_ptr<Enemy> enemy_);
    // Returns false on game-over.
    bool doDoSteps(FloatSeconds const &frame_time);
    int getScore() const;
    std::vector<std::unique_ptr<Projectile>>& getProjectileVector();
    DynamicPixelBuffer const& draw();
private:
    int levelWidth;
    int levelHeigth;

    DynamicPixelBuffer levelBuffer;

    Player& player;
    std::vector<Wall> wallVector;
    std::vector<Wall> borders;
    std::vector<Scenery> sceneryVector;
    std::vector<std::unique_ptr<Enemy>> enemyVector;
    std::vector<std::unique_ptr<Projectile>> projectileVector;
    std::vector<std::unique_ptr<Enemy>> trooperSpawner;
    bool collHelper(Position const& vecPos, PixelBuffer const& vecBuf, float buffPosX, float buffPosY, int buffWidth, int buffHeigth);

    int score = 0; //hauerch71498
};

#endif // LEVEL_HPP
