#ifndef PIXELBUFFER_HPP
#define PIXELBUFFER_HPP

#include <array>
#include <cstdint>
#include <optional>
#include <vector>

#include "Position.hpp"

enum class ColorValue : uint8_t {
	_0, _1, _2, _3, _4, _5, _6, _7, _8, _9, _A, _B, _C, _D, _E, _F, _t
};

struct Color {
	ColorValue v = ColorValue::_t;

	std::optional<uint_fast8_t> to_value() const {
		if(v == ColorValue::_t) {
			return std::nullopt;
		}
		return static_cast<uint_fast8_t>(v);
	}

	bool operator ==(Color const &other) const {
		return (v == other.v);
	}

	bool operator !=(Color const &other) const {
		return (v != other.v);
	}
};

// Until C++20 `using enum` is readily available...
static constexpr Color _0 = { ColorValue::_0 };
static constexpr Color _1 = { ColorValue::_1 };
static constexpr Color _2 = { ColorValue::_2 };
static constexpr Color _3 = { ColorValue::_3 };
static constexpr Color _4 = { ColorValue::_4 };
static constexpr Color _5 = { ColorValue::_5 };
static constexpr Color _6 = { ColorValue::_6 };
static constexpr Color _7 = { ColorValue::_7 };
static constexpr Color _8 = { ColorValue::_8 };
static constexpr Color _9 = { ColorValue::_9 };
static constexpr Color _A = { ColorValue::_A };
static constexpr Color _B = { ColorValue::_B };
static constexpr Color _C = { ColorValue::_C };
static constexpr Color _D = { ColorValue::_D };
static constexpr Color _E = { ColorValue::_E };
static constexpr Color _F = { ColorValue::_F };
static constexpr Color _t = { ColorValue::_t };

// Forward-declare the iterator, as it's needed in PixelBuffer, which is in
// turn needed by PixelIterator...
class PixelIterator;

// Actually more of a pixel shader interface rather than a buffer... Might
// have been inspired by Go's Image interface:
// https://golang.org/pkg/image/#Image
class PixelBuffer {
	friend class PixelIterator;

private:
	// Why does C++ not have destructive moves?! :( These could be `const` and
	// `public` otherwise.
	int width;
	int height;

protected:
	// Trap pixel, returned by peek() if [pos] is outside the bounds of the
	// image.
	static Color trap;

	// Safe version of peek(), implemented by derived classes. [pos] is
	// guaranteed to be ≥ (0, 0) and < (width, height).
	virtual Color peek_unchecked(PixelPoint const &pos) const = 0;

public:
	int const w() const { return width; }
	int const h() const { return height; }

	// Bounds-checked accessor.
	Color peek(PixelPoint const &pos) const;

	constexpr PixelBuffer(int w, int h)
		: width(w), height(h) {
	}

	PixelIterator begin() const;
	PixelIterator end() const;
};

// May not have been worth the over-engineering effort, given that it "just"
// flattens two nested `for` loop into a single one… but look, we can also get
// rid of the bounds checks this way! Then again, it's read-only… oh well.
class PixelIterator {
protected:
	PixelBuffer const *const buffer;
	PixelPoint position;

public:
	PixelIterator(PixelBuffer const &buffer, PixelPoint const &pos)
		: buffer(&buffer), position(pos) {
	}

	PixelPoint const& pos() const { return position; }
	Color peek() const            { return buffer->peek_unchecked(position); }

	void operator ++() {
		position.x++;
		if(position.x >= buffer->w()) {
			position.x = 0;
			position.y++;
		}
	}

	PixelIterator& operator *() {
		return *this;
	}

	bool operator !=(const PixelIterator &other) {
		return ((buffer != other.buffer) || (position != other.position));
	}
};

// A PixelBuffer with a std::array<W, H> as its buffer. Can be used if the
// size is known at compile time. Has a constexpr constructor, making it
// especially useful for sprite definitions.
template <size_t W, size_t H> class StaticPixelBuffer : public PixelBuffer {
public:
	// Since we take string literals with an implicit \0 at the end, the input
	// type has to have 1 more column.
	using BufferType      = std::array<std::array<Color, (W    )>, H>;
	using BufferInputType = std::array<std::array<char,  (W + 1)>, H>;

	BufferType const buffer;

	constexpr StaticPixelBuffer(BufferInputType const &chars)
		: PixelBuffer(W, H), buffer(filter_null_terminators(chars)) {
	}

	Color peek_unchecked(PixelPoint const &pos) const {
		return buffer[pos.y][pos.x];
	};

protected:
	static constexpr BufferType filter_null_terminators(
		BufferInputType const &input
	) {
		BufferType ret;
		// We can't use algorithms in constexpr functions anyway.
		for(int y = 0; y < H; ++y) {
			for(int x = 0; x < W; ++x) {
				switch(input[y][x]) {
				case ' ': ret[y][x] = _t; break;
				case '0': ret[y][x] = _0; break;
				case '1': ret[y][x] = _1; break;
				case '2': ret[y][x] = _2; break;
				case '3': ret[y][x] = _3; break;
				case '4': ret[y][x] = _4; break;
				case '5': ret[y][x] = _5; break;
				case '6': ret[y][x] = _6; break;
				case '7': ret[y][x] = _7; break;
				case '8': ret[y][x] = _8; break;
				case '9': ret[y][x] = _9; break;
				case 'A': ret[y][x] = _A; break;
				case 'B': ret[y][x] = _B; break;
				case 'C': ret[y][x] = _C; break;
				case 'D': ret[y][x] = _D; break;
				case 'E': ret[y][x] = _E; break;
				case 'F': ret[y][x] = _F; break;
				default:
					throw "Invalid color character";
				}
			}
		}
		return ret;
	}
};

// A PixelBuffer with a std::vector<Color> as its buffer. Also has unique
// blitting methods for copying the contents of other PixelBuffers to certain
// positions within this buffer.
class DynamicPixelBuffer : public PixelBuffer {
public:
	DynamicPixelBuffer(int w, int h);

	Color peek_unchecked(PixelPoint const &pos) const;
	// Returns [trap] if [pos] is outside the buffer boundaries.
	Color& at(PixelPoint const &pos);
	void clear(Color ch = {});

	void blit_topleft(PixelPoint const &topleft, PixelBuffer const &other);
	void blit_centered(PixelPoint const &center, PixelBuffer const &other);

protected:
	std::vector<Color> buffer;
};

// Base class for any pixel shader that takes an existing PixelBuffer as its
// input.
// …yeah, it does nothing other than enforce that structure. Why did we end up
// with this class? Eh, maybe it's nice for semantic reasons. Maybe.
// (Prof. Teßmann would disagree)
class ShadedPixelBuffer : public PixelBuffer {
protected:
	PixelBuffer const& input;

public:
	ShadedPixelBuffer(PixelBuffer const& input)
		: PixelBuffer(input.w(), input.h()), input(input) {
	}
};

#endif // PIXELBUFFER_HPP
